# pylint: disable-import-outside-toplevel
"""Netport fastapi server.

Contains all netport functionalities.
"""
import os
import re
import uuid
import socket
import platform
import subprocess
from os.path import exists

import psutil
from loguru import logger
from fastapi import FastAPI, Request
from redis import exceptions as redis_errors

from netport.common import R_PORT, R_PATH, R_PROCESS, Response
from netport.database import RedisDatabase, LocalDatabase, IDatabase

app = FastAPI()
running_processes = []
db: IDatabase = NotImplemented


def _is_port_in_use(port: int):
    """Checks if the given port is in use.

    Args:
        port (int): The port to check

    Returns:
        bool. If the port is in use.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


@app.get("/db/reserve")
def reserve(request: Request, resource: str, value):
    result = db.reserve(client_ip=request.client.host, resource=resource, value=value)
    return Response(data=result)


@app.get("/db/is_reserved")
def is_reserved(resource: str, value: str):
    """Check if the given resource is in use.

    Args:
        resource (str): The resource type
        value (str): The value of the resource

    Returns:
        bool. If the port is in use.
    """
    return Response(data=db.is_reserved(resource, value))


@app.get("/db/my_resources")
def my_resources(request: Request, resource: str):
    """Get all resources that are reserved to the client."""
    return Response(data=db.get_client_resources(request.client.host, resource))


@app.get("/db/get_client_resources")
def get_client_resources(
    request: Request,
    client_ip: str = None,
    resource: str = r"(.)+",
    value: str = r"(.)+",
):
    """Get all resources that are reserved for a specific client."""
    if not client_ip:
        client_ip = request.client.host

    return Response(data=db.get_client_resources(client_ip, resource, value))


@app.get("/db/release_resource")
def release_resource(request: Request, resource: str, value):
    """Release a resource for the requesting client."""
    return Response(data=db.release_resource(request.client.host, resource, value) == 1)


@app.get("/db/release_client")
def release_client(request: Request, client: str = None):
    """Release all resources that are reserved for the requesting client."""
    if client:
        return db.release_client(client) == 1

    return Response(db.release_client(request.client.host) == 1)


@app.get("/db/get_all_resources")
def get_all_resources():
    """Get all clients that hold resources."""
    return Response(data=db.get_all_resources())


@app.get("/db/get_all_clients")
def get_all_clients():
    """Get all clients that hold resources."""
    return Response(data=db.get_all_clients())


@app.get("/networking/get_port")
def get_port(request: Request, port: int = None):
    """Get an empty port.

    Opens a socket on random port, gets its port number, and closes the socket.

    This action is not atomic, so race condition is possible...
    """
    if port and not _is_port_in_use(port):
        if db.reserve(request.client.host, R_PORT, port):
            return Response(data={"port": port})

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        _, port = s.getsockname()
        if db.reserve(request.client.host, R_PORT, port):
            return Response(data={"port": port})

    return Response(data=False)


@app.get("/networking/is_port_in_use")
def is_port_in_use(port: int):
    """Checks if the given port is in use.

    Args:
        port (int): The port to use

    Returns:
        bool. If the port is in use.
    """
    return Response(data=_is_port_in_use(port))


@app.get("/networking/whats_my_ip")
def whats_my_ip(request: Request):
    """Check how netport sees the client's ip."""
    return Response(data=request.client.host)


@app.get("/networking/list_interfaces")
def list_interfaces():
    """List all available network interfaces.

    Returns:
        list. Network interfaces.
    """
    interfaces = psutil.net_if_addrs()
    return Response(data=list(interfaces.keys()))


@app.get("/fs/is_path_exist")
def is_path_exist(path: str):
    """Check if the given path exists on the machine."""
    return Response(data=exists(path.strip()))


@app.get("/fs/reserve_path")
def reserve_path(request: Request, path: str):
    """Reserve a path for the client."""
    if not exists(path):
        return Response(data=False)

    return Response(data=db.reserve(request.client.host, R_PATH, path))


@app.get("/shell/execute_command")
def execute_command(command: str):
    """Run a shell command and returns its result."""
    return Response(
        data=subprocess.call(
            command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        == 0
    )


@app.get("/shell/run_process")
def run_process(request: Request, command: str):
    """Run a process on the machine."""
    if db.reserve(request.client.host, R_PROCESS, command):
        process = subprocess.Popen(command.split(" "))
        running_processes.append(process)
        return Response(data=process.pid)

    return Response(data=-1)


@app.get("/shell/is_process_running")
def is_process_running(name: str):
    """Check if the process already runs."""
    for proc in psutil.process_iter():
        try:
            if name.lower() in proc.name().lower():
                return Response(data=proc.name())

        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass

    return Response(data=False)


@app.get("/system/get_system_info")
def get_system_info():
    return Response(
        data={
            "platform": platform.system(),
            "platform-release": platform.release(),
            "platform-version": platform.version(),
            "architecture": platform.machine(),
            "hostname": socket.gethostname(),
            "ip-address": socket.gethostbyname(socket.gethostname()),
            "mac-address": ":".join(re.findall("..", "%012x" % uuid.getnode())),
            "processor": platform.processor(),
            "ram": str(round(psutil.virtual_memory().total / (1024.0**3))) + " GB",
        }
    )


def run_app():
    global db
    redis_host = os.environ.get("NETPORT_REDIS_HOST", None)
    redis_port = os.environ.get("NETPORT_REDIS_PORT", None)
    redis_db = os.environ.get("NETPORT_REDIS_DB", None)

    try:
        if None in [redis_host, redis_port, redis_db]:
            raise AttributeError

        db = RedisDatabase(host=redis_host, port=redis_port, db_instance=redis_db)
        logger.success(
            f"Connected to Redis database at {redis_host}:{redis_port},{redis_db}"
        )

    except AttributeError:
        db = LocalDatabase()
        logger.success("Initiated a local database")

    except (redis_errors.ConnectionError, TypeError) as error:
        msg = (
            f"Couldn't connect to the redis database due to bad parameters: host={redis_host}, "
            f"port={redis_port}, db={redis_db}."
        )

        logger.error(msg)
        raise ConnectionError(msg) from error

    return app


if __name__ == "__main__":
    run_app()



from typing import Any
from dataclasses import dataclass

R_PORT = "PORT"
R_PATH = "PATH"
R_PROCESS = "PROCESS"


@dataclass
class Response:
    """Response format for every api command.

    Every response from the netport server will have the save stracture:
    {
        "data": <Response>
    }
    """

    data: Any



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['netport']

package_data = \
{'': ['*']}

install_requires = \
['docopt>=0.6.2,<0.7.0',
 'fastapi>=0.85.0,<0.86.0',
 'loguru>=0.6.0,<0.7.0',
 'psutil>=5.9.2,<6.0.0',
 'redis>=4.3.4,<5.0.0',
 'requests>=2.28.1,<3.0.0',
 'uvicorn[standard]>=0.18.3,<0.19.0']

entry_points = \
{'console_scripts': ['netport = scripts.cli:main']}

setup_kwargs = {
    'name': 'netport',
    'version': '0.4.0',
    'description': 'Tool for managing resources on a remove machine using openapi',
    'long_description': '# Netport\n\n![Licence](https://badgen.net/pypi/license/netport)\n![Python Versions](https://badgen.net/pypi/python/netport)\n![Pypi Version](https://badgen.net/pypi/v/netport)\n![Code Quality Status](https://github.com/IgalKolihman/netport/actions/workflows/pull_request.yaml/badge.svg)\n\nNetport is a tool for managing single-access resources on the target Unix machine. Netport manages\nthe access to different types of resources on the operating system by not allowing multiple requests\nto the same resource. For example ports, files, processes, network interfaces, and more...\n\n# How it works\n\n```mermaid\nflowchart TD\n    client([Client])\n    server[/Netport Server\\]\n    database[(Database)]\n    resources{{Resources}}\n    ports(Ports)\n    interfaces(Interfaces)\n    fs(File System)\n    proc(System Processes)\n    info(System Info)\n    \n    client <--REST API requests--> server\n    server <--stores server state---> database\n    server <----> resources\n    resources --- ports\n    resources --- interfaces\n    resources --- fs\n    resources --- proc\n    resources --- info\n```\n\n## Framework\n\nNetport runs in a single python process on the target machine. It uses the\n[FastAPI](https://fastapi.tiangolo.com/) framework to make a high performance and easy to use REST\napi server. By using this api, users can easily perform a variety of requests to various resources.\n\nThere are many types of requests that can be made:\n\n* Acquire a free port.\n* Check if file exists\n* Declare that a file is being used\n* Start a process\n* Get a list of already acquired resources\n\nAnd many more...\n\n## Backend Storage\n\nIn order to maintain an active memory of the used resources, Netport communicates with a database.\nThere are 2 types of supported databases that netport uses:\n\n* Redis database\n* Local pythonic database.\n\nBoth databases serve the same purpose for netport, but their inner functionality still a bit\ndifferent, with one draw back for the local database. **The local database doesn\'t save netport\'s\nstate after a shutdown or a reboot**.\n\nThe decision whether to use redis DB or a local one, depends by the user. In case of an unexpected\nshutdown, with redis the state of netport will be stored and on reboot it will continue from where\nit stopped. On the other hand, for much simpler systems this feature might be unnecessary\ncomplicated.\n\n# Installation\n\nMake sure that python is installed on your machine (3.7 and above). Open your terminal and run the\nfollowing command:\n\n```shell\npip install netport\n```\n\n> _It is advised to use a dedicated python virtual environment for netport._\n\n# Developer Installation\n\nInstall poetry by following the instructions [here](https://python-poetry.org/docs/).\n\n1. Clone this repository:\n\n```shell\ngit clone https://github.com/IgalKolihman/netport.git\n```\n\n2. Install the development environment:\n\n```shell\npoetry install --with dev\n```\n\n# Usage\n\nPlease follow the [installation procedure](#installation) to be able to run Netport. After that\nnetport will be available to you as long as you are using it\'s virtual environment.\n\nTo run netport with basic configurations, run:\n\n```shell\nnetport\n```\n\nAfter executing this command, a link will appear in the terminal to the server\'s url. The API\ndocumentation will be available at: _**"http://host_ip:port/docs"**_\n\nFor more advanced information regarding netport execution, run the following command:\n\n```shell\nnetport -h\n```',
    'author': 'Igal Kolihman',
    'author_email': 'igalk.spam@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/IgalKolihman/netport',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Running netport from the cli

Usage:
    netport -h | --help
    netport -g [<path>]
    netport -p <port>
    netport [-H <host>] [--redis=<redis-address>]
    netport [-H <host> -p <port>] [--redis=<redis-address>]

Options:
  -h --help             Show this screen.
  --redis=<redis-address>   Address and port to connect to the redis DB.
                        The correct format is: "<ip addr>:<port>:<db>"
  -H --host             IP address to bind the app to [default: 0.0.0.0].
  -p --port             Port to bind the app to [default: 80].
  -g                    Generate netport openapi scheme in a json file.

Examples:
    netport -H 127.0.0.1 -p 1234 --redis=192.168.1.125:6379:0
    netport --host 127.0.0.1 --port 1234 --redis=192.168.1.125:6379:0
    netport -g api.json
    netport -g /home/user/api.json
"""
import os
import json

import uvicorn
from docopt import docopt

DEFAULT_HOST = "0.0.0.0"
DEFAULT_PORT = 80


def generate_open_api_scheme(scheme_path: None):
    from netport.netport import app

    if scheme_path is None:
        scheme_path = "netport_openapi.json"

    scheme = app.openapi()
    with open(scheme_path, "w") as scheme_file:
        json.dump(scheme, scheme_file)

    print(f"The Netport scheme was exported to {scheme_path}")


def main():
    args = docopt(__doc__)
    if args["-g"]:
        generate_open_api_scheme(args["<path>"])
        exit()

    print(args)

    host = args["<host>"] if args["--host"] else DEFAULT_HOST
    port = int(args["<port>"]) if args["--port"] else DEFAULT_PORT

    if args["--redis"]:
        redis_addr, redis_port, redis_db = args["--redis"].split(":")
        os.environ["NETPORT_REDIS_HOST"] = redis_addr
        os.environ["NETPORT_REDIS_PORT"] = redis_port
        os.environ["NETPORT_REDIS_DB"] = redis_db

    from netport.netport import run_app

    uvicorn.run(run_app, host=host, port=port, log_level="info", factory=True)


if __name__ == "__main__":
    main()



"""Netport integration with different databases."""
import re
import time
import datetime

from redis.client import Redis
from redis.exceptions import ResponseError
from redis import exceptions as redis_errors


class IDatabase:
    """General interface that describes the functionality needed for in a netport database."""

    def is_reserved(self, resource: str, value):
        """Check if the requested resource is already reserved."""
        raise NotImplementedError

    def reserve(self, client_ip: str, resource: str, value) -> bool:
        """Reserve some resource for a client."""
        raise NotImplementedError

    def get_client_resources(
        self, client_ip: str, resource_regex: str = "*", value_regex: str = "*"
    ):
        """Get all the client resources that match the resource regex pattern."""
        raise NotImplementedError

    def release_resource(self, client_ip: str, resource: str, value):
        """Free the resource from its client."""
        raise NotImplementedError

    def release_client(self, client_ip: str):
        """Free the requested client from his resources."""
        raise NotImplementedError

    def get_all_clients(self):
        """Get all clients that posses resources in Netport."""
        raise NotImplementedError

    def get_all_resources(self):
        all_resources = {}
        for client in self.get_all_clients():
            client_resources = []
            for resource in self.get_client_resources(client):
                # Each resource might contain metadata about it in different databases. Therefore
                # to keep a similar usability of the different databases the `resource` is converted
                # to list before returning to the client.
                if type(resource) is list:
                    client_resources.append(resource)

                else:
                    client_resources.append(
                        [
                            resource,
                        ]
                    )

            all_resources[client] = client_resources

        return all_resources


class RedisDatabase(IDatabase):
    """Manages Netport's requests in a redis database.

    The data is saved and manages in Redis on a big hash table. Here is an example of one:

    | Client IP     | Resources                                 |
    | ------------- | ----------------------------------------- |
    | 192.168.0.128 | PORT:1234                                 |
    | 192.168.0.128 | PATH:/home/user/Downloads/version.tag.gz  |
    | 192.168.0.128 | PROCESS:ping 192.168.0.65                 |
    | 192.168.0.65  | PROCESS:bash /home/user/sniff_ping.sh     |

    In the given example we can see that there are 2 clients that currently use netport, client
    192.168.0.128 and 192.168.0.65. Also client 192.168.0.128 have requested 3 different resources
    compared to 192.168.0.65 who only requested 1 resource.
    """

    def __init__(self, host: str = "localhost", port: str = 6379, db_instance=0):
        try:
            self._db = Redis(host=host, port=int(port), db=db_instance)
            self._db.ping()

        except redis_errors.TimeoutError:
            raise TimeoutError("Couldn't connect to the Redis database")

    def is_reserved(self, resource: str, value):
        """Check if the requested resource is already reserved."""
        for client in self.get_all_clients():
            try:
                for _ in self.get_client_resources(client, resource, value):
                    return True
            except ResponseError as redis_error:
                if (
                    "WRONGTYPE Operation against a key holding the wrong kind of value"
                    in redis_error.args[0]
                ):
                    raise MemoryError(
                        "Seems like the Redis database have some keys that do not belong to "
                        f"Netport. Found the key '{client}'. Before continuing remove those types "
                        f"of keys."
                    ) from redis_error

                raise redis_error

        return False

    def reserve(self, client_ip: str, resource: str, value) -> bool:
        """Reserve some resource for a client."""
        if self.is_reserved(resource, value):
            return False

        return (
            self._db.hset(
                client_ip,
                f"{resource}:{value}",
                datetime.datetime.utcfromtimestamp(time.time()).strftime(
                    "%Y-%m-%d %H:%M:%S"
                ),
            )
            == 1
        )

    def get_client_resources(
        self, client_ip: str, resource_regex: str = "*", value_regex: str = "*"
    ):
        """Get all the client resources that match the resource regex pattern."""
        return self._db.hscan_iter(client_ip, match=f"{resource_regex}:{value_regex}")

    def release_resource(self, client_ip: str, resource: str, value):
        return self._db.hdel(client_ip, f"{resource}:{value}")

    def release_client(self, client_ip: str):
        return self._db.delete(client_ip) == 1

    def get_all_clients(self):
        return self._db.scan(cursor=0, match="*")[1]


class LocalDatabase(IDatabase):
    """Manages Netport's data in a pythonic dictionary without depending of any external tool.

    The data is saved inside a dictionary in which each key is a client containing a list of it's
    requested resources. The following is an example of how the data would look inside the
    LocalDatabase:

    {
        "192.168.0.128": [
            "PORT:1234",
            "PATH:/home/user/Downloads/version.tag.gz",
            "PROCESS:ping 192.168.0.65"
        ]
        "192.168.0.65" [
            "PROCESS:bash /home/user/sniff_ping.sh"
        ]
    }

    In the example above we can see the 2 clients, 192.168.0.128 and 192.168.0.65 with their
    resources.
    """

    def __init__(self):
        self._data = {}

    def is_reserved(self, resource: str, value):
        """Check if the requested resource is already reserved by some client."""
        record = rf"{resource}:{value}"
        for client in self._data.keys():
            if record in self._data[client]:
                return True

        return False

    def reserve(self, client_ip: str, resource: str, value) -> bool:
        """Reserve some resource for a client."""
        if self.is_reserved(resource, value):
            return False

        record = f"{resource}:{value}"
        try:
            self._data[client_ip].append(record)
        except KeyError:
            self._data[client_ip] = [
                record,
            ]

        return True

    def get_client_resources(
        self, client_ip: str, resource_regex: str = r"(.)+", value_regex: str = r"(.)+"
    ):
        """Get all the client resources that match the resource regex pattern."""
        if client_ip not in list(self._data.keys()):
            return []

        matched_resources = []

        for record in self._data[client_ip]:
            if re.search(rf"{resource_regex}:{value_regex}", record):
                matched_resources.append(record)

        return matched_resources

    def release_resource(self, client_ip: str, resource: str, value):
        record = f"{resource}:{value}"
        if record in self._data[client_ip]:
            self._data[client_ip].remove(record)
            return True

        return False

    def release_client(self, client_ip: str):
        try:
            self._data.pop(client_ip)
            return True
        except KeyError:
            return False

    def get_all_clients(self):
        return list(self._data.keys())
